"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const EventEmitter = require("events");
const kafka_node_1 = require("kafka-node");
const Nats = require("nats");
const util_1 = require("util");
const worker_threads_1 = require("worker_threads");
const logs_1 = require("./logs");
const logs_threads_1 = require("./logs_threads");
const threads_1 = require("./threads");
const types_1 = require("./types");
const util_2 = require("./util");
/**
 *  Clase para las conexiones y publicaciones sobre
 *  la cola de mensajes.
 */
class Brokers extends util_2.default {
    /**
     * Contructor de la clase.
     * @param kafkaHost Servidores Kafka
     * @param Topic Topico de escucha
     */
    constructor(Conf) {
        //  Contructor padre
        super();
        /**
         * Manejador de eventos
         * @protected
         */
        this.event = new EventEmitter();
        /**
         * Manejador de logs
         * @protected
         */
        this.logs = logs_1.default;
        /**
         * Servidores de kafka
         * @protected
         */
        this.kafkaHost = [];
        /**
         * Hosts de los brokers
         * @protected
         */
        this.Hosts = [];
        /**
         * Variable para revisar si NATS se encuentra conectado o desconectado
         * @protected
         */
        this.natsIsConnected = false;
        /**
         * variable que almacena la direccion
         * actual del archivo que se ejecuta.
         * @private
         */
        this.cwd = '';
        /**
         * Indicador si el hilo en reserva se encuentra activo.
         * @private
         */
        this.threadResInit = false;
        /**
         * Instancia de la conexion a NATS
         * @private
         */
        this.connNats = null;
        /**
         * PID de la conexion a la subscripcion en
         * NATS
         * @private
         */
        this.pidSub = null;
        /**
         * Numero de hilos registrados actualmente
         * @private
         */
        this.threadTotal = 1;
        /**
         * Maximo numero de hilos que se soportan
         * @private
         */
        this.maxThread = 0;
        /**
         * Indicador para conocer si el consumer se encuentra en pause.
         * @private
         */
        this.isPauseConsumer = false;
        /**
         * Indicador del soporte de hilos
         * @private
         */
        this.threadSupport = true;
        // Asignar la direccion actual
        this.cwd = process.cwd();
        // Verificar el soporte de hilos
        this.threadSupport = !util_1.isUndefined(Conf.Thread) ? Conf.Thread : false;
        // Asignar el numero maximo de hilos disponibles
        this.maxThread = process.env.MAX_THREAD ? parseInt(process.env.MAX_THREAD, 0) : 100;
        // Crear el administrador de los Thread
        this.thread = new threads_1.default(process.cwd(), __dirname);
        // Asignar la configuracion
        this.conf = Conf;
        this.conf.ClearQueque = util_1.isUndefined(Conf.ClearQueque) ? true : Conf.ClearQueque;
        /**
         * Definir topico con soporte para servicios con nomemclatura
         * legada.
         */
        if (this.conf.Version === 'v1' && this.conf.App === 'appcoppel') {
            // Es un servicio legado de appcoppel
            this.Topic = this.conf.Name;
        }
        else {
            // Crear la combinacion
            this.Topic = `${this.conf.App}_${this.conf.Version}_${this.conf.Name}`;
        }
        // Ver si corresponde a una conexion sobre NATS
        if (worker_threads_1.isMainThread) {
            // Validar si es NATS
            if (process.env.NATS) {
                this.connecNats();
            }
            else {
                // Kafka por defualt
                this.connectKafka();
            }
            if (this.threadSupport) {
                // Crear hilo
                this.threadRes = this.thread.execute({});
                this.threadRes.on('online', () => {
                    this.threadResInit = true;
                });
            }
        }
        else {
            // Es un hilo secundario
            worker_threads_1.parentPort.on('message', (data) => {
                // Publicar para su procesamiento
                this.message(data.Data);
            });
            // Cambiar el objeto de logs
            this.logs = new logs_threads_1.default();
        }
    }
    /**
     * Metodo para validar conexion al broker
     */
    validateConnection() {
        if (worker_threads_1.isMainThread) {
            if (process.env.NATS) {
                return this.natsIsConnected;
            }
            else {
                if (!this.client) {
                    return false;
                }
                return new Promise((rr, ff) => {
                    this.client.topicExists([this.Topic], (error) => {
                        rr(((error) ? false : true));
                    });
                });
            }
        }
        return false;
    }
    /**
     * Metodo para realizar el envio de datos a la cola de mensajes
     * @param data Datos a enviar
     * @param topic Topico de publicacion
     * @param key Clave unica
     */
    send(dataMessage, topic, key = util_2.default.key()) {
        if (worker_threads_1.isMainThread) {
            if (process.env.NATS) {
                // Responder la peticion
                this.nc.publish(topic, dataMessage);
            }
            else {
                this.producer.send([{
                        key,
                        messages: [dataMessage],
                        topic,
                    }], (err) => {
                    if (err) {
                        // Imprimir el error
                        this.logs.error(err);
                    }
                });
            }
        }
        else {
            // Publicar en el hilo principal
            worker_threads_1.parentPort.postMessage({
                Data: {
                    dataMessage,
                    key,
                    topic,
                },
                Type: types_1.TypeMessage.Publish,
            });
        }
    }
    /**
     * Metodo para realizar la conexion Nats
     * @private
     */
    connecNats() {
        // Verificar si se paso la cadena de conexion a kafka
        if (this.conf.Nats) {
            // Asignar la pasada como configuracion
            this.Hosts = this.conf.Nats;
        }
        else {
            // Verificar que se asigno una por variable de ambiente
            if (process.env.NATS) {
                // Recorrer todos los elementos para darle formato
                for (const i of process.env.NATS.split(',')) {
                    this.Hosts.push(`nats://${i}`); // Dar el formato correcto
                }
            }
            else {
                this.logs.error('No se proporciono la configuracion de NATS');
                process.exit(-1); // Finalizar el programa
            }
        }
        // Realizar la conexion a NATS - nc.currentServer.url.host
        this.nc = Nats.connect({ servers: this.Hosts, name: this.Topic });
        // Evento que indica la conexion
        this.nc.on('connect', (nc) => {
            const text = `Escuchando el topico ${this.Topic} en Nats ${this.nc.currentServer.url.host}`;
            // Notificar que la conexion fue exitosa.
            this.logs.info(text);
            // Almacenar la conexion
            this.connNats = nc;
            // Realizar la conexion de subcripcion al topico
            this.pidSub = this.connNats.subscribe(this.Topic, { queue: this.Topic }, this.handlerNats.bind(this));
            this.natsIsConnected = true;
        });
        this.nc.on('disconnect', () => {
            this.natsIsConnected = false;
        });
        this.nc.on('reconnect', () => {
            this.natsIsConnected = true;
        });
    }
    /**
     * Metodo para realizar la conexion al broker de kafka
     * @private
     */
    connectKafka() {
        // Verificar si se paso la cadena de conexion a kafka
        if (this.conf.Broker) {
            // Asignar la pasada como configuracion
            this.kafkaHost = this.conf.Broker;
        }
        else {
            // Verificar que se asigno una por variable de ambiente
            if (process.env.KAFKA) {
                // Asignar
                this.kafkaHost = process.env.KAFKA.split(',');
            }
            else {
                this.logs.error('No se proporciono la configuracion de KAFKA');
                process.exit(-1); // Finalizar el programa
            }
        }
        // Crear el cliente de kafka.
        this.client = new kafka_node_1.KafkaClient({ kafkaHost: this.kafkaHost.join(',') });
        // Registrar el producer.
        this.producer = new kafka_node_1.Producer(this.client);
        // Eventos del producer
        this.producer.on('ready', () => {
            // Opciones del consumidor
            const options = {
                groupId: this.Topic,
                kafkaHost: this.kafkaHost.join(','),
            };
            // Crear el consumidor
            this.consumer = new kafka_node_1.ConsumerGroup(options, [this.Topic]);
            // Escuchar cuando se realice la conexion
            this.consumer.on('connect', () => {
                const text = `Escuchando el topico ${this.Topic} en el Kafka ${this.kafkaHost.join(',')}`;
                // Notificar que la conexion fue exitosa.
                this.logs.info(text);
            });
            // Llegada de mensajes
            this.consumer.on('message', (message) => {
                try {
                    // Convertir el mensaje
                    const data = JSON.parse(Buffer.from(message.value.toString(), 'base64').toString('utf8'));
                    // Pasar el mensaje al que lo procesara
                    this.processMessage(data);
                }
                catch (error) {
                    // Disparamos la excepcion
                    this.exception(this.Topic, message.value.toString(), error);
                }
            });
            // Errores
            this.consumer.on('error', (err) => {
                // Enviar el error
                this.logs.error(err);
            });
        });
        this.producer.on('error', (err) => {
            // Enviar el error
            this.logs.error(err);
        });
    }
    /**
     * Handler para la escueha de mensajes de NATS
     * @param msg Mensaje de entrada
     * @param reply Id del que solicita el dato
     * @param subject Topico
     */
    handlerNats(msg, reply, subject) {
        try {
            // Convertir el mensaje
            let data = JSON.parse(Buffer.from(msg, 'base64').toString('utf8'));
            // agregar el ID para el reply
            if (!data.uuid && !data.metadata.asynchronous) {
                data = Object.assign(data, { uuid: reply });
            }
            // Pasar el mensaje al que lo procesara
            this.processMessage(data);
        }
        catch (error) {
            // Disparamos la excepcion
            this.exception(this.Topic, msg, error);
        }
    }
    /**
     * Metodo para crear una instancia del servicio para
     * el procesamiento del mensaje.
     * @param data Datos que se envian a la instancia
     */
    processMessage(data) {
        // Validar el soporte de hilos
        if (this.threadSupport) {
            const w = this.threadRes;
            this.threadRes = this.thread.execute(data);
            // Agregar un elemento
            this.threadTotal++;
            // Verificar si se pausa el consumer
            if (!this.isPauseConsumer && this.threadTotal >= (this.maxThread - 1)) {
                // Iniciar de nuevo el consumer
                this.pause();
            }
            // Evento de conexion
            this.threadRes.on('online', () => {
                this.threadResInit = true;
            });
            // Validar si el hilo ya se encuentra activo
            if (this.threadResInit) {
                w.postMessage({ type: 'message', Data: data });
            }
            else {
                w.on('online', () => {
                    w.postMessage({ type: 'message', Data: data });
                });
            }
            // Esperar los mensajes
            w.on('message', (response) => {
                if (response.Type === types_1.TypeMessage.Publish) {
                    const dataResponse = response.Data;
                    // Publicar
                    this.send(dataResponse.dataMessage, dataResponse.topic, dataResponse.key);
                    // Eliminar el Hilo
                    w.terminate();
                    // Restar un elemento
                    this.threadTotal--;
                    // Ver si se requiere reanudar el consumer
                    if (this.isPauseConsumer && this.threadTotal < (this.maxThread - 1)) {
                        // Iniciar de nuevo el consumer
                        this.resume();
                    }
                }
                else if (response.Type === types_1.TypeMessage.Logs) {
                    const logs = response.Data;
                    if (logs.Type === types_1.TypeLogs.Info) {
                        this.logs.info(logs.Message);
                    }
                    else if (logs.Type === types_1.TypeLogs.Error) {
                        this.logs.error(logs.Message);
                    }
                    else if (logs.Type === types_1.TypeLogs.Warning) {
                        this.logs.warning(logs.Message);
                    }
                    else if (logs.Type === types_1.TypeLogs.Log) {
                        this.logs.log(logs.Message);
                    }
                }
            });
            w.on('error', (error) => {
                this.logs.error(error);
            });
            w.on('exit', (exitCode) => {
                this.logs.info(`Worker ${w.threadId} has stopped with code ${exitCode}`);
            });
        }
        else {
            this.message(data);
        }
    }
    /**
     * Metodo para pausar la llegada de mensajes
     * desde el broker
     * @private
     */
    pause() {
        if (this.conf.Nats) {
            this.connNats.drainSubscription(this.pidSub);
        }
        else {
            this.consumer.pause();
        }
        // Indicar que se esta en pause
        this.isPauseConsumer = true;
    }
    /**
     * Metodo para activar la escucha de mensajes
     * desde el broker.
     * @private
     */
    resume() {
        if (this.conf.Nats) {
            this.pidSub = this.connNats.subscribe(this.Topic, { queue: this.Topic }, this.handlerNats.bind(this));
        }
        else {
            this.consumer.resume();
        }
        // Indicar que sale de la pausa
        this.isPauseConsumer = false;
    }
}
exports.default = Brokers;
